
/*
 * CINELERRA
 * Copyright (C) 2008 Adam Williams <broadcast at earthling dot net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include "audiodevice.h"
#include "audioalsa.h"
#include "bcsignals.h"
#include "mainerror.h"
#include "mutex.h"
#include "playbackconfig.h"
#include "preferences.h"
#include "recordconfig.h"

#include <errno.h>

#ifdef HAVE_ALSA

#define DEFAULT_DEVICE "default"

AudioALSA::AudioALSA(AudioDevice *device)
 : AudioLowLevel(device)
{
	samples_written = 0;
	delay = 0;
	timing_lock = new Mutex("AudioALSA::timing_lock");
	interrupted = 0;
	dsp_out = 0;
}

AudioALSA::~AudioALSA()
{
	delete timing_lock;
}

void AudioALSA::list_devices(ArrayList<char*> *devices, int pcm_title, int mode)
{
	snd_ctl_t *handle;
	int card, err, dev, idx;
	snd_ctl_card_info_t *info;
	snd_pcm_info_t *pcminfo;
	char string[BCTEXTLEN];
	snd_pcm_stream_t stream = SND_PCM_STREAM_PLAYBACK;
	int error;
	switch(mode)
	{
	case MODERECORD:
		stream = SND_PCM_STREAM_CAPTURE;
		break;
	case MODEPLAY:
		stream = SND_PCM_STREAM_PLAYBACK;
		break;
	}

	snd_ctl_card_info_alloca(&info);
	snd_pcm_info_alloca(&pcminfo);

	card = -1;
	char *result = new char[strlen(DEFAULT_DEVICE) + 1];
	devices->append(result);
	devices->set_array_delete();     // since we are allocating by new[]
	strcpy(result, DEFAULT_DEVICE);

	while(snd_card_next(&card) >= 0)
	{
		char name[BCTEXTLEN];
		if(card < 0) break;
		sprintf(name, "hw:%i", card);

		if((err = snd_ctl_open(&handle, name, 0)) < 0)
		{
			errorbox("Failed to open ALSA card=%i: %s", card, snd_strerror(err));
			continue;
		}

		if((err = snd_ctl_card_info(handle, info)) < 0)
		{
			errorbox("Failed to get ALSA info card=%i: %s", card, snd_strerror(err));
			snd_ctl_close(handle);
			continue;
		}

		dev = -1;

		while(1)
		{
			unsigned int count;
			if(snd_ctl_pcm_next_device(handle, &dev) < 0)
				errorbox("Failed to get next ALSA device");

			if (dev < 0)
				break;

			snd_pcm_info_set_device(pcminfo, dev);
			snd_pcm_info_set_subdevice(pcminfo, 0);
			snd_pcm_info_set_stream(pcminfo, stream);

			if((err = snd_ctl_pcm_info(handle, pcminfo)) < 0) 
			{
				if(err != -ENOENT)
					errorbox("ALSA card=%i: %s\n", card, snd_strerror(err));
				continue;
			}

			if(pcm_title)
			{
				sprintf(string, "plughw:%d,%d", card, dev);
			}
			else
			{
				sprintf(string, "%s #%d", 
					snd_ctl_card_info_get_name(info), 
					dev);
			}

			char *result = devices->append(new char[strlen(string) + 1]);
			strcpy(result, string);
		}

		snd_ctl_close(handle);
	}
}

void AudioALSA::translate_name(char *output, char *input)
{
	ArrayList<char*> titles;
	ArrayList<char*> pcm_titles;
	int mode;

	if(device->r) mode = MODERECORD;
	else
	if(device->w) mode = MODEPLAY;

	list_devices(&titles, 0, mode);
	list_devices(&pcm_titles, 1, mode);

	sprintf(output, "default");
	for(int i = 0; i < titles.total; i++)
	{
		if(!strcasecmp(titles.values[i], input))
		{
			strcpy(output, pcm_titles.values[i]);
			break;
		}
	}

	titles.remove_all_objects();
	pcm_titles.remove_all_objects();
}

snd_pcm_format_t AudioALSA::translate_format(int format)
{
	switch(format)
	{
	case 8:
		return SND_PCM_FORMAT_S8;

	case 16:
		return SND_PCM_FORMAT_S16_LE;

	case 24:
		return SND_PCM_FORMAT_S24_LE;

	case 32:
		return SND_PCM_FORMAT_S32_LE;
	}
	return SND_PCM_FORMAT_UNKNOWN;
}

int AudioALSA::set_params(snd_pcm_t *dsp, 
	int channels, 
	int bits,
	int samplerate,
	int samples)
{
	snd_pcm_hw_params_t *params;
	snd_pcm_sw_params_t *swparams;

	snd_pcm_hw_params_alloca(&params);
	snd_pcm_sw_params_alloca(&swparams);

	if(snd_pcm_hw_params_any(dsp, params) < 0)
	{
		errormsg("ALSA: no PCM configurations available");
		return 1;
	}

	if(snd_pcm_hw_params_set_access(dsp, 
		params,
		SND_PCM_ACCESS_RW_INTERLEAVED))
	{
		errormsg("ALSA: failed to set up interleaved device access.");
		return 1;
	}

	if(snd_pcm_hw_params_set_format(dsp, 
		params, 
		translate_format(bits)))
	{
		errormsg("ALSA: failed to set output format.");
		return 1;
	}

	if(snd_pcm_hw_params_set_channels(dsp, 
		params, 
		channels))
	{
		errormsg("ALSA: Configured ALSA device does not support %d channel operation.", channels);
		return 1;
	}

	if(snd_pcm_hw_params_set_rate_near(dsp, 
		params, 
		(unsigned int*)&samplerate, 
		(int*)0))
	{
		errormsg("ALSA: Configured ALSA device does not support %u Hz playback.", samplerate);
		return 1;
	}

// Buffers written must be equal to period_time
	int buffer_time;
	int period_time;
	if(device->r)
	{
		buffer_time = 10000000;
		period_time = (int)((int64_t)samples * 1000000 / samplerate);
	}
	else
	{
		buffer_time = (int)((int64_t)samples * 1000000 * 4 / samplerate);
		period_time = buffer_time / 2;
		sleep_delay = period_time / 8;
	}
	snd_pcm_hw_params_set_buffer_time_near(dsp, 
		params,
		(unsigned int*)&buffer_time, 
		(int*)0);
	snd_pcm_hw_params_set_period_time_near(dsp, 
		params,
		(unsigned int*)&period_time, 
		(int*)0);
	if(snd_pcm_hw_params(dsp, params) < 0)
	{
		errormsg("Failed to set ALSA hw_params");
		return 1;
	}
	snd_pcm_sw_params_current(dsp, swparams);

	snd_pcm_sw_params_set_avail_min(dsp, swparams, samples);
	if(snd_pcm_sw_params(dsp, swparams) < 0)
	{
		errorbox("Failed to set ALSA sw_params");
		/* we can continue staggering along even if this fails */
	}

	device->device_buffer = samples * bits / 8 * channels;
	return 0;
}

int AudioALSA::open_input()
{
	char pcm_name[BCTEXTLEN];
	snd_pcm_stream_t stream = SND_PCM_STREAM_CAPTURE;
	int open_mode = 0;
	int err;

	device->in_channels = device->get_ichannels();
	device->in_bits = device->in_config->alsa_in_bits;

	translate_name(pcm_name, device->in_config->alsa_in_device);

	if(err = snd_pcm_open(&dsp_in, pcm_name, stream, open_mode))
	{
		dsp_in = 0;
		errorbox("Failed to open ALSA input: %s\n", snd_strerror(err));
		return 1;
	}

	if(set_params(dsp_in, 
		device->get_ichannels(), 
		device->in_config->alsa_in_bits,
		device->in_samplerate,
		device->in_samples))
	{
		errormsg("ALSA: open_input: set_params failed. Aborting sampling.");
		close_input();
		return 1;
	}

	return 0;
}

int AudioALSA::open_output()
{
	char pcm_name[BCTEXTLEN];
	snd_pcm_stream_t stream = SND_PCM_STREAM_PLAYBACK;
	int open_mode = 0;
	int err;

	device->out_channels = device->get_ochannels();
	device->out_bits = device->out_config->alsa_out_bits;

	translate_name(pcm_name, device->out_config->alsa_out_device);

	if((err = snd_pcm_open(&dsp_out, pcm_name, stream, open_mode)) < 0)
	{
		dsp_out = 0;
		errorbox("Failed to open ALSA output %s: %s", pcm_name, snd_strerror(err));
		return 1;
	}

	if(set_params(dsp_out, 
		device->get_ochannels(), 
		device->out_config->alsa_out_bits,
		device->out_samplerate,
		device->out_samples))
	{
		errormsg("ALSA: open_output: set_params failed.  Aborting playback.");
		close_output();
		return 1;
	}
	return 0;
}

int AudioALSA::open_duplex()
{
// ALSA always opens 2 devices
	return 0;
}

void AudioALSA::close_output()
{
	if(device->w && dsp_out)
	{
		snd_pcm_close(dsp_out);
		dsp_out = 0;
	}
}

void AudioALSA::close_input()
{
	if(device->r)
	{
		snd_pcm_drop(dsp_in);
		snd_pcm_drain(dsp_in);
		snd_pcm_close(dsp_in);
		dsp_in = 0;
	}
}

void AudioALSA::close_all()
{
	close_input();
	close_output();
	if(device->d)
	{
		snd_pcm_close(dsp_duplex);
	}
	samples_written = 0;
	delay = 0;
	interrupted = 0;
}

samplenum AudioALSA::device_position()
{
	snd_pcm_sframes_t delay;
	samplenum result;

	timing_lock->lock("AudioALSA::device_position");
	snd_pcm_delay(get_output(), &delay);
	result = samples_written - delay;
	timing_lock->unlock();
	return result;
}

int AudioALSA::read_buffer(char *buffer, int size)
{
	int attempts = 0;
	int done = 0;

	if(!get_input())
	{
		sleep(1);
		return 0;
	}

	while(attempts < 1 && !done)
	{
		if(snd_pcm_readi(get_input(), 
			buffer, 
			size / (device->in_bits / 8) / device->get_ichannels()) < 0)
		{
			printf("AudioALSA::read_buffer overrun at sample %lld\n", 
				device->total_samples_read);
			close_input();
			open_input();
			attempts++;
		}
		else
			done = 1;
	}
	return 0;
}

int AudioALSA::write_buffer(char *buffer, int size)
{
// Don't give up and drop the buffer on the first error.
	int attempts = 0;
	snd_pcm_sframes_t samples = size / (device->out_bits / 8) / device->get_ochannels();
	snd_pcm_sframes_t written, avail;
	int rc;

	if(!get_output()) return 0;
// Wait until enough buffer frees
// We want to avoid snd_pcm_writei blocking
	snd_pcm_wait(get_output(), -1);
	while((avail = snd_pcm_avail(get_output())) < samples)
	{
		usleep(sleep_delay);
	}

	timing_lock->lock("AudioALSA::write_buffer");
	while(attempts < 2 && !interrupted)
	{
		if((written = snd_pcm_writei(get_output(), 
			buffer, 
			samples)) < 0)
		{
			if(written == -EPIPE)
			{
// Underrun
				if((rc = snd_pcm_prepare(get_output())) < 0 && attempts == 0)
				{
					errorbox("ALSA write underrun at %.3f: %s. Recovery failed.",
						device->current_postime(), snd_strerror(rc));
				}
				attempts++;
			}
			else
			{
				timing_lock->unlock();
				errorbox("ALSA write failed: %s", snd_strerror(written));
				return -1;
			}
		}
		else
		{
			samples_written += written;
			break;
		}
	}
	timing_lock->unlock();

	if(attempts >= 2)
		return -1;
// Give time to video decoding
	usleep(sleep_delay);
	return 0;
}

void AudioALSA::flush_device()
{
	if(get_output())
		snd_pcm_drain(get_output());
}

void AudioALSA::interrupt_playback()
{
	if(get_output()) 
	{
		interrupted = 1;
// Interrupts the playback but may not have caused snd_pcm_writei to exit.
// With some soundcards it causes snd_pcm_writei to freeze for a few seconds.
// Use lock to avoid snd_pcm_drop called during snd_pcm_writei
		timing_lock->lock("AudioALSA::interrupt_playback");
		snd_pcm_drop(get_output());
// Prepare is needed to end snd_pcm_drain
		snd_pcm_prepare(get_output());
		timing_lock->unlock();
	}
}


snd_pcm_t* AudioALSA::get_output()
{
	if(device->w) return dsp_out;
	else
	if(device->d) return dsp_duplex;
	return 0;
}

snd_pcm_t* AudioALSA::get_input()
{
	if(device->r) return dsp_in;
	else
	if(device->d) return dsp_duplex;
	return 0;
}

#endif
