
/*
 * CINELERRA
 * Copyright (C) 2008 Adam Williams <broadcast at earthling dot net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include "asset.h"
#include "bcsignals.h"
#include "brender.h"
#include "clip.h"
#include "condition.h"
#include "edl.h"
#include "edlsession.h"
#include "language.h"
#include "mainsession.h"
#include "mainerror.h"
#include "mtimebar.h"
#include "mutex.h"
#include "mwindowgui.h"
#include "mwindow.h"
#include "packagedispatcher.h"
#include "preferences.h"
#include "renderfarm.h"
#include "tracks.h"
#include "units.h"

#include <errno.h>
#include <signal.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>

extern "C"
{
#include <uuid/uuid.h>
}


BRender::BRender(MWindow *mwindow)
 : Thread()
{
	this->mwindow = mwindow;
	map_lock = new Mutex("BRender::map_lock");
	completion_lock = new Condition(0, "BRender::completion_lock");
	timer = new Timer;
	socket_path[0] = 0;
	thread = 0;
	master_pid = -1;
	arguments[0] = arguments[1] = arguments[2] = 0;
	map_valid = 0;
	set_synchronous(1);
}

BRender::~BRender()
{
	if(thread) 
	{
		stop();
		delete thread;
	}


	if(master_pid >= 0)
	{
		kill(master_pid, SIGKILL);
		Thread::join();
	}

	delete map_lock;
	delete completion_lock;
	UNSET_TEMP(socket_path);
	remove(socket_path);
	if(arguments[0]) delete [] arguments[0];
	if(arguments[1]) delete [] arguments[1];
	if(arguments[2]) delete [] arguments[2];
	delete timer;
}

void BRender::initialize()
{
	timer->update();
// Create socket for background process.
	uuid_t socket_temp;
	sprintf(socket_path, "/tmp/cinelerra.");
	uuid_generate(socket_temp);
	uuid_unparse(socket_temp, socket_path + strlen(socket_path));
	SET_TEMP(socket_path);

// Start background instance of executable since codecs aren't reentrant
	Thread::start();

// Wait for local node to start
	thread = new BRenderThread(mwindow, this);
	thread->initialize();
}

void BRender::run()
{
	char string[BCTEXTLEN];
	int size;
	FILE *fd;

// Construct executable command with the designated filesystem port
	fd = fopen("/proc/self/cmdline", "r");
	if(fd)
	{
		if(fread(string, 1, BCTEXTLEN, fd) <= 0)
			perror("BRender::fork_background: can't read /proc/self/cmdline.\n");
		fclose(fd);
	}
	else
		perror(_("BRender::fork_background: can't open /proc/self/cmdline.\n"));

	arguments[0] = new char[strlen(string) + 1];
	strcpy(arguments[0], string);

	strcpy(string, "-b");
	arguments[1] = new char[strlen(string) + 1];
	strcpy(arguments[1], string);

	arguments[2] = new char[strlen(socket_path) + 1];
	strcpy(arguments[2], socket_path);

	arguments[3] = 0;

	int pid = vfork();
	if(!pid)
	{
		execvp(arguments[0], arguments);
		perror("BRender::fork_background");
		_exit(0);
	}

	master_pid = pid;

	int return_value;
	if(waitpid(master_pid, &return_value, WUNTRACED) < 0)
	{
		perror("BRender::run waitpid");
	}
}

// Give the last position of the EDL which hasn't changed.
// We copy the EDL and restart rendering at the lesser of position and
// our position.
void BRender::restart(EDL *edl)
{
	BRenderCommand *new_command = new BRenderCommand;
	map_valid = 0;
	new_command->copy_edl(edl);
	new_command->command = BRenderCommand::BRENDER_RESTART;
	thread->send_command(new_command);
// Map should be reallocated before this returns.
}

void BRender::stop()
{
	BRenderCommand *new_command = new BRenderCommand;
	new_command->command = BRenderCommand::BRENDER_STOP;
	thread->send_command(new_command);
	completion_lock->lock("BRender::stop");
}

void BRender::allocate_map(ptstime brender_start, ptstime start, ptstime end)
{
	map_lock->lock("BRender::allocate_map");

	videomap.set_map(0, brender_start, 0);
	videomap.set_map(start, end, 0);
	map_valid = 1;
	mwindow->session->brender_end = start;
	map_lock->unlock();
}

void BRender::set_video_map(ptstime start, ptstime end)
{
	int update_gui = 0;
	map_lock->lock("BRender::set_video_map");

	if(start < 0)
		start = 0;
	if(end < 0)
		end = 0;
	if(PTSEQU(start, end))
		return;
	videomap.set_map(start, end, 1);

// Maintain last contiguous here to reduce search time
	if(videomap.last->pts > mwindow->session->brender_end)
	{
		mwindow->session->brender_end = videomap.last->pts;

		if(timer->get_difference() > 1000)
		{
			update_gui = 1;
			timer->update();
		}
	}

	map_lock->unlock();

	if(update_gui)
	{
		mwindow->gui->lock_window("BRender::set_video_map");
		mwindow->gui->timebar->update(1, 0);
		mwindow->gui->timebar->flush();
		mwindow->gui->unlock_window();
	}
}


BRenderCommand::BRenderCommand()
{
	edl = 0;
	command = BRENDER_NONE;
	position = 0.0;
}

BRenderCommand::~BRenderCommand()
{
// EDL should be zeroed if copied
	if(edl) delete edl;
}

void BRenderCommand::copy_from(BRenderCommand *src)
{
	this->edl = src->edl;
	src->edl = 0;
	this->position = src->position;
	this->command = src->command;
}


void BRenderCommand::copy_edl(EDL *edl)
{
	this->edl = new EDL;
	this->edl->create_objects();
	this->edl->copy_all(edl);
	this->position = 0;
}





BRenderThread::BRenderThread(MWindow *mwindow, BRender *brender)
 : Thread(1)
{
	this->mwindow = mwindow;
	this->brender = brender;
	input_lock = new Condition(0, "BRenderThread::input_lock");
	thread_lock = new Mutex("BRenderThread::thread_lock");
	command_queue = 0;
	command = 0;
	done = 0;
	farm_server = 0;
	farm_result = 0;
	preferences = 0;
}

BRenderThread::~BRenderThread()
{
	thread_lock->lock("BRenderThread::~BRenderThread");
	done = 1;
	input_lock->unlock();
	thread_lock->unlock();
	Thread::join();
	delete input_lock;
	delete thread_lock;
	if(command) delete command;
	if(command_queue) delete command_queue;
	if(preferences) delete preferences;
}


void BRenderThread::initialize()
{
	Thread::start();
}

void BRenderThread::send_command(BRenderCommand *command)
{
	thread_lock->lock("BRenderThread::send_command");

	if(this->command_queue)
	{
		delete this->command_queue;
		this->command_queue = 0;
	}
	this->command_queue = command;

	input_lock->unlock();
	thread_lock->unlock();
}

int BRenderThread::is_done(int do_lock)
{
	if(do_lock) thread_lock->lock("BRenderThread::is_done");
	int result = done;
	if(do_lock) thread_lock->unlock();
	return result;
}

void BRenderThread::run()
{
	while(!is_done(1))
	{
		BRenderCommand *new_command = 0;
		thread_lock->lock("BRenderThread::run 1");

// Got new command
		if(command_queue)
		{
			;
		}
		else
// Wait for new command
		{
			thread_lock->unlock();
			input_lock->lock("BRenderThread::run 2");
			thread_lock->lock("BRenderThread::run 3");
		}

// Pull the command off
		if(!is_done(0))
		{
			new_command = command_queue;
			command_queue = 0;
		}

		thread_lock->unlock();

// Process the command here to avoid delay.
// Quit condition
		if(!new_command)
		{
			;
		}
		else
		if(new_command->command == BRenderCommand::BRENDER_STOP)
		{
			stop();
			delete new_command;
			new_command = 0;
		}
		else
		if(new_command->command == BRenderCommand::BRENDER_RESTART)
		{
// Compare EDL's and get last equivalent position in new EDL
			if(command && command->edl)
				new_command->position = 
					new_command->edl->equivalent_output(command->edl);
			else
				new_command->position = 0;


			stop();
			brender->completion_lock->lock("BRenderThread::run 4");

			if(new_command->edl->tracks->total_playable_vtracks())
			{
				if(command) delete command;
				command = new_command;
				start();
			}
			else
			{
				delete new_command;
				new_command = 0;
			}
		}
	}
}

void BRenderThread::stop()
{
	if(farm_server)
	{
		farm_result = 1;
		farm_server->wait_clients();
		delete farm_server;
		delete packages;
		delete preferences;
		farm_server = 0;
		packages = 0;
		preferences = 0;
	}
	brender->completion_lock->unlock();
}

void BRenderThread::start()
{
// Reset return parameters
	farm_result = 0;
	fps_result = 0;
	int result = 0;

// Allocate render farm.
	if(!farm_server)
	{
		preferences = new Preferences;
		preferences->copy_from(mwindow->preferences);
		packages = new PackageDispatcher;

// Fix preferences to use local node
		if(!preferences->use_renderfarm)
		{
			preferences->use_renderfarm = 1;
			preferences->delete_nodes();
		}
		preferences->add_node(brender->socket_path,
			0,
			1,
			preferences->local_rate);
		preferences->brender_asset->use_header = 0;
		preferences->brender_asset->frame_rate = command->edl->session->frame_rate;
		preferences->brender_asset->width = command->edl->session->output_w;
		preferences->brender_asset->height = command->edl->session->output_h;
		preferences->brender_asset->interlace_mode = command->edl->session->interlace_mode;

// Get last contiguous and reset map.
// If the framerate changes, last good should be 0 from the user.
		ptstime brender_start = command->edl->session->brender_start;
		ptstime last_contiguous = brender->videomap.last->pts;
		ptstime last_good = command->position;
		if(last_good < 0) last_good = last_contiguous;
		ptstime start_pts = MIN(last_contiguous, last_good);
		start_pts = MAX(start_pts, brender_start);
		ptstime end_pts = command->edl->tracks->total_video_length();
		if(end_pts < start_pts) end_pts = start_pts;

		brender->allocate_map(brender_start, start_pts, end_pts);

		result = packages->create_packages(mwindow,
			command->edl,
			preferences,
			BRENDER_FARM, 
			preferences->brender_asset, 
			start_pts,
			end_pts,
			0);

		farm_server = new RenderFarmServer(mwindow->plugindb, 
			packages,
			preferences,
			0,
			&farm_result,
			0,
			0,
			preferences->brender_asset,
			command->edl,
			brender);

		result = farm_server->start_clients();

// No local rendering because of codec problems.

// Abort
		if(result)
		{
// No-one must be retrieving a package when packages are deleted.
			delete farm_server;
			delete packages;
			delete preferences;
			farm_server = 0;
			packages = 0;
			preferences = 0;
		}
	}
}
