
/*
 * CINELERRA
 * Copyright (C) 2008 Adam Williams <broadcast at earthling dot net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include "aframe.h"
#include "amodule.h"
#include "atrack.h"
#include "attachmentpoint.h"
#include "autoconf.h"
#include "bcsignals.h"
#include "cplayback.h"
#include "cwindow.h"
#include "edl.h"
#include "edlsession.h"
#include "floatautos.h"
#include "localsession.h"
#include "mainerror.h"
#include "mainprogress.h"
#include "mainundo.h"
#include "menueffects.h"
#include "mwindow.h"
#include "mwindowgui.h"
#include "playbackengine.h"
#include "plugin.h"
#include "pluginaclient.h"
#include "pluginaclientlad.h"
#include "pluginclient.h"
#include "pluginserver.h"
#include "pluginvclient.h"
#include "preferences.h"
#include "sema.h"
#include "mainsession.h"
#include "trackcanvas.h"
#include "transportque.h"
#include "vdevicex11.h"
#include "vframe.h"
#include "videodevice.h"
#include "virtualanode.h"
#include "virtualvnode.h"
#include "vmodule.h"
#include "vtrack.h"


#include <sys/types.h>
#include <sys/wait.h>
#include <dlfcn.h>


PluginServer::PluginServer()
{
	reset_parameters();
	modules = new ArrayList<Module*>;
	nodes = new ArrayList<VirtualNode*>;
}

PluginServer::PluginServer(const char *path)
{
	reset_parameters();
	set_path(path);
	modules = new ArrayList<Module*>;
	nodes = new ArrayList<VirtualNode*>;
}

PluginServer::PluginServer(PluginServer &that)
{
	reset_parameters();

	if(that.title)
	{
		title = new char[strlen(that.title) + 1];
		strcpy(title, that.title);
	}

	if(that.path)
	{
		path = new char[strlen(that.path) + 1];
		strcpy(path, that.path);
	}

	modules = new ArrayList<Module*>;
	nodes = new ArrayList<VirtualNode*>;

	attachment = that.attachment;
	realtime = that.realtime;
	multichannel = that.multichannel;
	preferences = that.preferences;
	synthesis = that.synthesis;
	audio = that.audio;
	video = that.video;
	theme = that.theme;
	fileio = that.fileio;
	uses_gui = that.uses_gui;
	mwindow = that.mwindow;
	keyframe = that.keyframe;
	plugin_fd = that.plugin_fd;
	new_plugin = that.new_plugin;

	is_lad = that.is_lad;
	lad_descriptor = that.lad_descriptor;
	lad_descriptor_function = that.lad_descriptor_function;
}

PluginServer::~PluginServer()
{
	close_plugin();
	if(path) delete [] path;
	if(title) delete [] title;
	if(modules) delete modules;
	if(nodes) delete nodes;
	if(picon) delete picon;
}

// Done only once at creation
int PluginServer::reset_parameters()
{
	mwindow = 0;
	keyframe = 0;
	prompt = 0;
	cleanup_plugin();
	plugin_fd = 0;
	autos = 0;
	plugin = 0;
	edl = 0;
	preferences = 0;
	title = 0;
	path = 0;
	audio = video = theme = 0;
	uses_gui = 0;
	realtime = multichannel = fileio = 0;
	synthesis = 0;
	start_auto = end_auto = 0;
	picon = 0;
	transition = 0;
	new_plugin = 0;
	client = 0;
	use_opengl = 0;
	vdevice = 0;

	is_lad = 0;
	lad_descriptor_function = 0;
	lad_descriptor = 0;
	for(int i = 0; i < AFRAMES_IN_PLUGIN; i++)
		aframes_used[i] = 0;
}


// Done every time the plugin is opened or closed
int PluginServer::cleanup_plugin()
{
	in_buffer_size = out_buffer_size = 0;
	total_in_buffers = total_out_buffers = 0;
	error_flag = 0;
	written_samples = 0;
	shared_buffers = 0;
	new_buffers = 0;
	written_samples = written_frames = 0;
	gui_on = 0;
	plugin = 0;
	plugin_open = 0;
}

void PluginServer::set_mwindow(MWindow *mwindow)
{
	this->mwindow = mwindow;
}

void PluginServer::set_attachmentpoint(AttachmentPoint *attachmentpoint)
{
	this->attachmentpoint = attachmentpoint;
}

void PluginServer::set_keyframe(KeyFrame *keyframe)
{
	this->keyframe = keyframe;
}

void PluginServer::set_prompt(MenuEffectPrompt *prompt)
{
	this->prompt = prompt;
}


int PluginServer::set_path(const char *path)
{
	if(this->path) delete [] this->path;
	this->path = new char[strlen(path) + 1];
	strcpy(this->path, path);
}

void PluginServer::set_title(const char *string)
{
	if(title) delete [] title;
	title = new char[strlen(string) + 1];
	strcpy(title, string);
}

void PluginServer::generate_display_title(char *string)
{
	if(plugin && plugin->track) 
		sprintf(string, "%s: %s", plugin->track->title, title);
	else
		strcpy(string, title);
}

// Open plugin for signal processing
int PluginServer::open_plugin(int master, 
	Preferences *preferences,
	EDL *edl, 
	Plugin *plugin,
	int lad_index)
{
	if(plugin_open) return 0;

	this->preferences = preferences;
	this->plugin = plugin;
	this->edl = edl;

	if(!new_plugin && !plugin_fd) plugin_fd = dlopen(path, RTLD_NOW);

	if(!new_plugin && !plugin_fd)
	{
// If the dlopen failed it may still be an executable tool for a specific
// file format, in which case we just store the path.
		set_title(path);
		char string[BCTEXTLEN];
		strcpy(string, dlerror());

		if(!strstr(string, "executable"))
			printf("PluginServer::open_plugin: %s\n", string);

		return 0;
	}


	if(!new_plugin && !lad_descriptor)
	{
		new_plugin = (PluginClient* (*)(PluginServer*))dlsym(plugin_fd, "new_plugin");

// Probably a LAD plugin but we're not going to instantiate it here anyway.
		if(!new_plugin)
		{
			lad_descriptor_function = (LADSPA_Descriptor_Function)dlsym(
				plugin_fd,
				"ladspa_descriptor");

			if(!lad_descriptor_function)
			{
// Not a recognized plugin
				fprintf(stderr, "PluginServer::open_plugin: new_plugin undefined in %s\n", path);
				dlclose(plugin_fd);
				plugin_fd = 0;
				return PLUGINSERVER_NOT_RECOGNIZED;
			}
			else
			{
// LAD plugin,  Load the descriptor and get parameters.
				is_lad = 1;
				if(lad_index >= 0)
				{
					lad_descriptor = lad_descriptor_function(lad_index);
				}

// make plugin initializer handle the subplugins in the LAD plugin or stop
// trying subplugins.
				if(!lad_descriptor)
				{
					dlclose(plugin_fd);
					plugin_fd = 0;
					return PLUGINSERVER_IS_LAD;
				}
			}
		}
	}


	if(is_lad)
	{
		client = new PluginAClientLAD(this);
	}
	else
	{
		client = new_plugin(this);
	}

	realtime = client->is_realtime();
	audio = client->is_audio();
	video = client->is_video();
	theme = client->is_theme();
	fileio = client->is_fileio();
	uses_gui = client->uses_gui();
	multichannel = client->is_multichannel();
	synthesis = client->is_synthesis();
	transition = client->is_transition();
	set_title(client->plugin_title());

	if(master)
	{
		picon = client->new_picon();
	}

	plugin_open = 1;
	return PLUGINSERVER_OK;
}

int PluginServer::close_plugin()
{
	if(!plugin_open) return 0;

	int plugin_status, result;
	if(client) delete client;

	plugin_open = 0;

	cleanup_plugin();
	return 0;
}

void PluginServer::client_side_close()
{
// Last command executed in client thread
	if(plugin)
		mwindow->hide_plugin(plugin, 1);
	else
	if(prompt)
	{
		prompt->lock_window("PluginServer::client_side_close");
		prompt->set_done(1);
		prompt->unlock_window();
	}
}

void PluginServer::render_stop()
{
	if(client)
		client->render_stop();
}


void PluginServer::init_realtime(int realtime_sched,
		int total_in_buffers, 
		int buffer_size)
{
	if(!plugin_open) return;

// set for realtime priority
// initialize plugin
// Call start_realtime
	this->total_in_buffers = this->total_out_buffers = total_in_buffers;
	client->plugin_init_realtime(realtime_sched, 
		total_in_buffers, 
		buffer_size);
}


// Replaced by pull method but still needed for transitions
void PluginServer::process_transition(VFrame *input,
		VFrame *output, 
		ptstime current_postime,
		ptstime total_len)
{
	if(!plugin_open) return;
	PluginVClient *vclient = (PluginVClient*)client;

	vclient->source_position = plugin->track->to_units(current_postime);
	vclient->source_start = 0;
	vclient->total_len = plugin->track->to_units(total_len);

	vclient->input = new VFrame*[1];
	vclient->output = new VFrame*[1];

	vclient->input[0] = input;
	vclient->output[0] = output;

	vclient->process_realtime(input, output);
	vclient->age_temp();
	delete [] vclient->input;
	delete [] vclient->output;
	use_opengl = 0;
}

void PluginServer::process_transition(AFrame *input, 
		AFrame *output,
		ptstime current_postime,
		ptstime total_len)
{
	if(!plugin_open) return;

	int offset = round((output->pts - input->pts) * output->samplerate);
	if(offset < 0)
		offset = 0;
	PluginAClient *aclient = (PluginAClient*)client;
	aclient->source_position = round(current_postime * output->samplerate);
	aclient->total_len = round(total_len * output->samplerate);
	aclient->source_start = 0;
	aclient->process_realtime(input->length,
		input->buffer,
		output->buffer + offset);
}


void PluginServer::process_buffer(VFrame **frame, 
	ptstime total_length)
{
	if(!plugin_open) return;
	PluginVClient *vclient = (PluginVClient*)client;
	double framerate;
	ptstime duration = frame[0]->get_duration();
	if(duration > EPSILON)
		framerate = 1.0 / duration;
	else
		framerate = edl->session->frame_rate;

	vclient->source_position = round(frame[0]->get_pts() * framerate);
	vclient->frame_rate = framerate;
	vclient->input = new VFrame*[total_in_buffers];
	vclient->output = new VFrame*[total_in_buffers];
	for(int i = 0; i < total_in_buffers; i++)
	{
		vclient->input[i] = frame[i];
		vclient->output[i] = frame[i];
	}
	if(plugin)
	{
		vclient->source_start = round(plugin->project_pts * framerate);
		vclient->total_len = round(total_length * framerate);
	}
	vclient->direction = PLAY_FORWARD;

	if(multichannel)
	{
		vclient->process_buffer(frame, vclient->source_position, framerate);
	}
	else
	{
		vclient->process_buffer(frame[0], vclient->source_position, framerate);
	}

	for(int i = 0; i < total_in_buffers; i++)
		frame[i]->push_prev_effect(title);

	delete [] vclient->input;
	delete [] vclient->output;

	vclient->age_temp();
	use_opengl = 0;
}

/*
 * Hack: mark AFrame as filled
 */
void PluginServer::mark_as_filled(AFrame *aframe)
{
	int l;
	if(aframe->length == 0)
	{
		l = aframe->length = aframe->source_length;
		aframe->duration = (ptstime)l / aframe->samplerate;
	}
}

void PluginServer::process_buffer(AFrame **buffer,
	ptstime total_len)
{
	if(!plugin_open) return;
	PluginAClient *aclient = (PluginAClient*)client;
	AFrame *aframe = buffer[0];
	int fragment_size;

	if(aframe->source_duration > 0)
		fragment_size = round(aframe->source_duration * aframe->samplerate);
	else
		fragment_size = aframe->source_length;

	if(aframe->samplerate <= 0)
		aframe->samplerate = aclient->project_sample_rate;

	aclient->source_position = round(aframe->pts * aframe->samplerate);
	aclient->source_pts = aframe->pts;
	aclient->sample_rate = aframe->samplerate;
	aclient->total_len_pts = total_len;

	if(plugin)
	{
		aclient->source_start = round(plugin->project_pts *
			aclient->project_sample_rate);
		aclient->source_start_pts = plugin->project_pts;
		aclient->total_len = round(total_len * aframe->samplerate);
	}

	aclient->direction = PLAY_FORWARD;
	if(aclient->has_pts_api())
	{
		if(multichannel)
		{
			for(int i = 1; i < total_in_buffers; i++)
			{
				buffer[i]->pts = aclient->source_pts;
				buffer[i]->source_length = fragment_size;
			}
			aclient->process_frame(buffer);
		}
		else
			aclient->process_frame(buffer[0]);
		return;
	}

	if(multichannel){
		double *samples[total_in_buffers];
		for(int i = 0; i < total_in_buffers; i++)
		{
			samples[i] = buffer[i]->buffer;
			put_aframe(buffer[i]);
		}
		aclient->process_buffer(fragment_size, 
			samples,
			aclient->source_position, 
			aclient->sample_rate);
		for(int i = 0; i < total_in_buffers; i++)
		{
			pop_aframe(buffer[i]);
			mark_as_filled(buffer[i]);
		}
	}
	else
	{
		put_aframe(buffer[0]);
		aclient->process_buffer(fragment_size, 
			buffer[0]->buffer, 
			aclient->source_position, 
			aclient->sample_rate);
// Hack - hope that buffer is filled by plugin
		pop_aframe(buffer[0]);
		mark_as_filled(buffer[0]);
	}
}


void PluginServer::send_render_gui(void *data)
{
	if(attachmentpoint) attachmentpoint->render_gui(data);
}

void PluginServer::send_render_gui(void *data, int size)
{
	if(attachmentpoint) attachmentpoint->render_gui(data, size);
}

void PluginServer::render_gui(void *data)
{
	if(client) client->plugin_render_gui(data);
}

void PluginServer::render_gui(void *data, int size)
{
	if(client) client->plugin_render_gui(data, size);
}

MainProgressBar* PluginServer::start_progress(char *string, int64_t length)
{
	mwindow->gui->lock_window("PluginServer::start_progress");
	MainProgressBar *result = mwindow->mainprogress->start_progress(string, length);
	mwindow->gui->unlock_window();
	return result;
}

samplenum PluginServer::get_written_samples()
{
	if(!plugin_open) return 0;
	return written_samples;
}

framenum PluginServer::get_written_frames()
{
	if(!plugin_open) return 0;
	return written_frames;
}


// ======================= Non-realtime plugin

int PluginServer::get_parameters(ptstime start, ptstime end, int channels)
{
	double rate;

	if(!plugin_open) return 0;

	if(video)
		rate = edl->session->frame_rate;
	else
		rate = edl->session->sample_rate;

	client->start = round(start * rate);
	client->end = round(end * rate);
	client->source_start = client->start;
	client->total_len = client->end - client->start;
	client->total_in_buffers = channels;
	return client->plugin_get_parameters();
}

int PluginServer::set_interactive()
{
	if(!plugin_open) return 0;
	client->set_interactive();
	return 0;
}

void PluginServer::append_module(Module *module)
{
	modules->append(module);
}

void PluginServer::append_node(VirtualNode *node)
{
	nodes->append(node);
}

void PluginServer::reset_nodes()
{
	nodes->remove_all();
}

int PluginServer::set_error()
{
	error_flag = 1;
	return 0;
}

int PluginServer::set_realtime_sched()
{
	struct sched_param params;
	params.sched_priority = 1;
	return 0;
}


int PluginServer::process_loop(VFrame **buffers, int &write_length)
{
	if(!plugin_open) return 1;
	return client->plugin_process_loop(buffers, write_length);
}

int PluginServer::process_loop(AFrame **buffers, int &write_length)
{
	double *samples[total_in_buffers];
	int result;

	if(!plugin_open) return 1;
	for(int i = 0; i < total_in_buffers; i++)
	{
		samples[i] = buffers[i]->buffer;
		put_aframe(buffers[i]);
	}
	result = client->plugin_process_loop(samples, write_length);
	for(int i = 0; i < total_in_buffers; i++)
	{
		pop_aframe(buffers[i]);
	}
	return result;
}


void PluginServer::start_loop(ptstime start,
	ptstime end,
	int buffer_size, 
	int total_buffers)
{
	double rate;

	if(!plugin_open) return;

	if(video)
		rate = edl->session->frame_rate;
	else
		rate = edl->session->sample_rate;
	total_in_buffers = total_buffers;
	client->plugin_start_loop(round(start * rate), round(end * rate),
		buffer_size, total_buffers);
	return;
}

void PluginServer::stop_loop()
{
	if(plugin_open)
		client->plugin_stop_loop();
}

void PluginServer::read_frame(VFrame *buffer, 
	int channel, 
	framenum start_position)
{
	buffer->set_pts((ptstime)start_position / mwindow->edl->session->frame_rate);
	buffer->set_layer(channel);
	((VModule*)modules->values[channel])->render(buffer,
		0,
		0);
}

void PluginServer::read_samples(double *buffer, 
	int channel,
	samplenum start_position, 
	int total_samples)
{
	AFrame *aframe = find_aframe(buffer);
	AFrame localframe;

	if(!aframe)
	{
		aframe = &localframe;
		aframe->set_buffer(buffer, total_samples);
	} else
		aframe->reset_buffer();
	if(aframe->samplerate == 0)
		aframe->samplerate = edl->session->sample_rate;
	aframe->channel = channel;
	aframe->pts = (ptstime)start_position/aframe->samplerate;
	aframe->source_length = total_samples;
	((AModule*)modules->values[aframe->channel])->render(aframe);
}


int PluginServer::read_frame(VFrame *buffer, 
	int channel, 
	framenum start_position,
	double frame_rate,
	int use_opengl)
{
// Data source depends on whether we're part of a virtual console or a
// plugin array.
//     VirtualNode
//     Module
// If we're a VirtualNode, read_data in the virtual plugin node handles
//     backward propogation and produces the data.
// If we're a Module, render in the module produces the data.
	int result = -1;
	if(!multichannel) channel = 0;

// Push our name on the next effect stack
	buffer->push_next_effect(title);
	buffer->set_pts((ptstime)start_position/frame_rate);
	buffer->set_layer(channel);
	if(nodes->total > channel)
	{
		result = ((VirtualVNode*)nodes->values[channel])->read_data(buffer,
			use_opengl);
	}
	else
	if(modules->total > channel)
	{
		result = ((VModule*)modules->values[channel])->render(buffer,
			0,
			use_opengl);
	}
	else
	{
		errorbox("PluginServer::read_frame no object available for channel=%d",
			channel);
	}

// Pop our name from the next effect stack
	buffer->pop_next_effect();

	return result;
}

int PluginServer::read_samples(double *buffer,
	int channel,
	int sample_rate,
	samplenum start_position, 
	int len)
{
	AFrame *aframe = find_aframe(buffer);
	AFrame localframe;

	if(!aframe){
		aframe = &localframe;
		aframe->set_buffer(buffer, len);
	}
	aframe->channel = channel;
	aframe->pts = (ptstime)start_position / sample_rate;
	aframe->source_length = len;
	aframe->source_pts = 0;
	aframe->samplerate = sample_rate;

	if(!multichannel) aframe->channel = 0;
	if(nodes->total > aframe->channel)
		return ((VirtualANode*)nodes->values[aframe->channel])->read_data(aframe);
	else
	if(modules->total > aframe->channel)
		return ((AModule*)modules->values[aframe->channel])->render(aframe);
	else
	{
		errorbox("PluginServer::read_samples no object available for channel=%d",
			aframe->channel);
	}
	return -1;
}

void PluginServer::get_aframe_rt(AFrame *aframe)
{
	if(!multichannel)
		aframe->channel = 0;
	if(nodes->total > aframe->channel)
	{
		((VirtualANode*)nodes->values[aframe->channel])->read_data(aframe);
		return;
	}
	if(modules->total > aframe->channel)
	{
		((AModule*)modules->values[aframe->channel])->render(aframe);
		return;
	}
	aframe->clear_buffer();
	if(aframe->buffer && aframe->source_length)
	{
		aframe->length = aframe->source_length;
		aframe->duration = round((ptstime)aframe->length / aframe->samplerate);
	}
}

void PluginServer::raise_window()
{
	if(!plugin_open) return;
	client->raise_window();
}

void PluginServer::show_gui()
{
	if(!plugin_open) return;
	client->smp = preferences->processors - 1;
	if(plugin)
	{
		client->total_len_pts = plugin->length();
		client->total_len = plugin->track->to_units(plugin->length());
		client->source_start_pts = plugin->project_pts;
		client->source_start = plugin->track->to_units(plugin->project_pts);
	}
	client->source_pts = mwindow->edl->local_session->get_selectionstart(1);
	if(video)
	{
		client->source_position = Units::to_int64(
			mwindow->edl->local_session->get_selectionstart(1) * 
				mwindow->edl->session->frame_rate);
	}
	else
	if(audio)
	{
		client->source_position = Units::to_int64(
			mwindow->edl->local_session->get_selectionstart(1) * 
				mwindow->edl->session->sample_rate);
	}
	client->update_display_title();
	client->show_gui();
}

void PluginServer::update_gui()
{
	if(!plugin_open || !plugin) return;

	client->total_len = plugin->track->to_units(plugin->length());
	client->total_len_pts = plugin->length();
	client->source_start = plugin->track->to_units(plugin->project_pts);
	client->source_start_pts = plugin->project_pts;
	client->source_pts = mwindow->edl->local_session->get_selectionstart(1);
	if(video)
	{
		client->source_position = Units::to_int64(
			mwindow->edl->local_session->get_selectionstart(1) * 
				mwindow->edl->session->frame_rate);
	}
	else
	if(audio)
	{
		client->source_position = Units::to_int64(
			mwindow->edl->local_session->get_selectionstart(1) * 
				mwindow->edl->session->sample_rate);
	}
	client->update_gui();
}

void PluginServer::update_title()
{
	if(!plugin_open) return;
	client->update_display_title();
}


int PluginServer::set_string(const char *string)
{
	if(!plugin_open) return 0;

	client->set_string_client(string);
	return 0;
}

int PluginServer::gui_open()
{
	if(attachmentpoint) return attachmentpoint->gui_open();
	return 0;
}

void PluginServer::set_use_opengl(int value, VideoDevice *vdevice)
{
	this->use_opengl = value;
	this->vdevice = vdevice;
}

int PluginServer::get_use_opengl()
{
	return use_opengl;
}


void PluginServer::run_opengl(PluginClient *plugin_client)
{
	if(vdevice)
		((VDeviceX11*)vdevice->get_output_base())->run_plugin(plugin_client);
}

// ============================= queries

int PluginServer::get_samplerate()
{
	if(!plugin_open) return 0;
	if(audio)
	{
		return client->get_samplerate();
	}
	else
	if(mwindow)
		return mwindow->edl->session->sample_rate;
	else
	{
		errorbox("PluginServer::get_samplerate audio and mwindow == NULL");
		return 1;
	}
}


double PluginServer::get_framerate()
{
	if(!plugin_open) return 0;
	if(video)
	{
		return client->get_framerate();
	}
	else
	if(mwindow)
		return mwindow->edl->session->frame_rate;
	else 
	{
		errorbox("PluginServer::get_framerate video and mwindow == NULL");
		return 1;
	}
}

int PluginServer::get_project_samplerate()
{
	if(mwindow)
		return mwindow->edl->session->sample_rate;
	else
	if(edl)
		return edl->session->sample_rate;
	else
	{
		errorbox("PluginServer::get_project_samplerate mwindow and edl are NULL.");
		return 1;
	}
}

double PluginServer::get_project_framerate()
{
	if(mwindow)
		return mwindow->edl->session->frame_rate;
	else
	if(edl)
		return edl->session->frame_rate;
	else
	{
		errorbox("PluginServer::get_project_framerate mwindow and edl are NULL.");
		return 1;
	}
}



void PluginServer::detach_buffers(void)
{
	ring_buffers_out.remove_all();
	offset_out_render.remove_all();
	double_buffer_out_render.remove_all();
	realtime_out_size.remove_all();

	ring_buffers_in.remove_all();
	offset_in_render.remove_all();
	double_buffer_in_render.remove_all();
	realtime_in_size.remove_all();

	out_buffer_size = 0;
	shared_buffers = 0;
	total_out_buffers = 0;
	in_buffer_size = 0;
	total_in_buffers = 0;
}

void PluginServer::arm_buffer(int buffer_number, 
		posnum offset_in, 
		posnum offset_out,
		int double_buffer_in,
		int double_buffer_out)
{
	offset_in_render.values[buffer_number] = offset_in;
	offset_out_render.values[buffer_number] = offset_out;
	double_buffer_in_render.values[buffer_number] = double_buffer_in;
	double_buffer_out_render.values[buffer_number] = double_buffer_out;
}


void PluginServer::set_automation(FloatAutos *autos, FloatAuto **start_auto, FloatAuto **end_auto, int reverse)
{
	this->autos = autos;
	this->start_auto = start_auto;
	this->end_auto = end_auto;
	this->reverse = reverse;
}



void PluginServer::save_data(KeyFrame *keyframe)
{
	if(!plugin_open) return;
	client->save_data(keyframe);
}

KeyFrame* PluginServer::get_prev_keyframe(posnum position)
{
	KeyFrame *result = 0;
	if(plugin)
		result = plugin->get_prev_keyframe(plugin->track->from_units(position));
	else
		result = keyframe;
	return result;
}

KeyFrame* PluginServer::prev_keyframe_pts(ptstime postime)
{
	if(plugin)
		return plugin->get_prev_keyframe(postime);
	return keyframe;
}

KeyFrame* PluginServer::get_next_keyframe(posnum position)
{
	KeyFrame *result = 0;
	if(plugin)
		result = plugin->get_next_keyframe(plugin->track->from_units(position));
	else
		result = keyframe;
	return result;
}

KeyFrame* PluginServer::next_keyframe_pts(ptstime postime)
{
	if(plugin)
		return plugin->get_next_keyframe(postime);
	return keyframe;
}

KeyFrame* PluginServer::get_keyframe()
{
	if(plugin)
		return plugin->get_keyframe();
	else
		return keyframe;
}

void PluginServer::get_camera(float *x, float *y, float *z,
	framenum position)
{
	plugin->track->automation->get_camera(x, y, z, 
		plugin->track->automation->pos2pts(position));
}

void PluginServer::get_projector(float *x, float *y, float *z,
	framenum position)
{
	plugin->track->automation->get_projector(x, y, z, 
		plugin->track->automation->pos2pts(position));
}


int PluginServer::get_interpolation_type()
{
	return plugin->edl->session->interpolation_type;
}

Theme* PluginServer::new_theme()
{
	if(theme)
	{
		return client->new_theme();
	}
	else
		return 0;
}

Theme* PluginServer::get_theme()
{
	if(mwindow) return mwindow->theme;
	return 0;
}


// Called when plugin interface is tweeked
void PluginServer::sync_parameters()
{
	if(video) mwindow->restart_brender();
	mwindow->sync_parameters();
	if(mwindow->edl->session->auto_conf->plugins)
	{
		mwindow->gui->lock_window("PluginServer::sync_parameters");
		mwindow->gui->canvas->draw_overlays();
		mwindow->gui->canvas->flash();
		mwindow->gui->unlock_window();
	}
}

int PluginServer::put_aframe(AFrame *af)
{
	for(int i = 0; i < AFRAMES_IN_PLUGIN; i++)
		if(aframes_used[i] == 0)
		{
			aframes_used[i] = af;
			return i;
		}
	errormsg("Out of aframes_used in '%s'", title);
	return -1;
}

AFrame* PluginServer::find_aframe(double *buffer)
{
	for(int i = 0; i < AFRAMES_IN_PLUGIN; i++)
		if(aframes_used[i] && aframes_used[i]->buffer == buffer)
		{
			return aframes_used[i];
		}
	return 0;
}

void PluginServer::pop_aframe(AFrame *af)
{
	for(int i = 0; i < AFRAMES_IN_PLUGIN; i++)
		if(aframes_used[i] && aframes_used[i] == af)
		{
			aframes_used[i] = 0;
			return;
		}
	errorbox("AFrame lost from pluginserver");
}

const char *PluginServer::plugin_conf_dir()
{
	if(edl)
		return edl->session->plugin_configuration_directory;
	return BCASTDIR;
}

void PluginServer::dump()
{
	printf("    PluginServer %s %s\n", path, title);
}
