
/*
 * CINELERRA
 * Copyright (C) 2008 Adam Williams <broadcast at earthling dot net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include "bcdisplayinfo.h"
#include "clip.h"
#include "bchash.h"
#include "delayaudio.h"
#include "filexml.h"
#include "language.h"
#include "picon_png.h"
#include "vframe.h"
#include <algorithm>
#include <string.h>


PluginClient* new_plugin(PluginServer *server)
{
	return new DelayAudio(server);
}

DelayAudio::DelayAudio(PluginServer *server)
 : PluginAClient(server)
{
	PLUGIN_CONSTRUCTOR_MACRO
}

DelayAudio::~DelayAudio()
{
	PLUGIN_DESTRUCTOR_MACRO
}


const char* DelayAudio::plugin_title() { return N_("Delay audio"); }
int DelayAudio::is_realtime() { return 1; }
int DelayAudio::has_pts_api() { return 1; }

int DelayAudio::load_configuration()
{
	KeyFrame *prev_keyframe;
	prev_keyframe = get_prev_keyframe(get_source_position());

	read_data(prev_keyframe);
	return 0;
}

void DelayAudio::load_defaults()
{
	char directory[BCTEXTLEN];

	plugin_configuration_path(directory, "delayaudio.rc");
	defaults = new BC_Hash(directory);
	defaults->load();
	config.length = defaults->get("LENGTH", (double)1);
}

void DelayAudio::save_defaults()
{
	defaults->update("LENGTH", config.length);
	defaults->save();
}

void DelayAudio::read_data(KeyFrame *keyframe)
{
	FileXML input;
	input.set_shared_string(keyframe->data, strlen(keyframe->data));

	int result = 0;
	while(!result)
	{
		result = input.read_tag();

		if(!result)
		{
			if(input.tag.title_is("DELAYAUDIO"))
			{
				config.length = input.tag.get_property("LENGTH", (double)config.length);
			}
		}
	}
}

void DelayAudio::save_data(KeyFrame *keyframe)
{
	FileXML output;
	output.set_shared_string(keyframe->data, MESSAGESIZE);

	output.tag.set_title("DELAYAUDIO");
	output.tag.set_property("LENGTH", (double)config.length);
	output.append_tag();
	output.tag.set_title("/DELAYAUDIO");
	output.append_tag();
	output.append_newline();
	output.terminate_string();
}

void DelayAudio::process_frame_realtime(AFrame *input, AFrame *output)
{
	int size = input->length;

	load_configuration();
	samplenum num_delayed = config.length * PluginAClient::project_sample_rate + 0.5;

	// Examples:
	//     buffer  size   num_delayed
	// A:    2       5        2        short delay
	// B:   10       5       10        long delay
	// C:    1       5       10        long delay, beginning
	// D:    1       5        2        short delay, beginning
	// E:   10       5        2        short delay after long delay

	samplenum num_silence = num_delayed - buffer.size();
	if (size < num_silence)
		num_silence = size;

	// Ex num_silence ==  A: 0, B: 0, C: 5, D: 1, E: -8

	buffer.insert(buffer.end(), input->buffer, input->buffer + size);

	// Ex buffer.size() ==  A: 7, B: 15, C: 6, D: 6, E: 15

	if(input != output)
		output->copy_of(input);

	if (num_silence > 0)
	{
		std::fill_n(output->buffer, num_silence, 0.0);
		size -= num_silence;
	}
	// Ex size ==  A: 5, B: 5, C: 0, D: 4, E: 5

	if (buffer.size() >= num_delayed + size)
	{
		std::vector<double>::iterator from = buffer.end() - (num_delayed + size);
		// usually, from == buffer.begin(); but if the delay has just been
		// reduced compared to the previous frame, then we drop samples

		// Ex from points to idx A: 0, B: 0, C: n/a, D: 0, E: 8

		std::copy(from, from + size, output->buffer);
		buffer.erase(buffer.begin(), from + size);
	}
}

SHOW_GUI_MACRO(DelayAudio, DelayAudioThread);
SET_STRING_MACRO(DelayAudio);
NEW_PICON_MACRO(DelayAudio);
RAISE_WINDOW_MACRO(DelayAudio);

void DelayAudio::update_gui()
{
	if(thread)
	{
		load_configuration();
		thread->window->lock_window();
		thread->window->update_gui();
		thread->window->unlock_window();
	}
}

PLUGIN_THREAD_OBJECT(DelayAudio, DelayAudioThread, DelayAudioWindow);

DelayAudioWindow::DelayAudioWindow(DelayAudio *plugin, int x, int y)
 : BC_Window(plugin->gui_string, 
	x, 
	y, 
	200, 
	80, 
	200, 
	80, 
	0, 
	0,
	1)
{
	this->plugin = plugin;
}

DelayAudioWindow::~DelayAudioWindow()
{
}

void DelayAudioWindow::create_objects()
{
	set_icon(new VFrame(picon_png));
	add_subwindow(new BC_Title(10, 10, _("Delay seconds:")));
	add_subwindow(length = new DelayAudioTextBox(plugin, 10, 40));
	update_gui();
	show_window();
	flush();
}


void DelayAudioWindow::update_gui()
{
	char string[BCTEXTLEN];

	sprintf(string, "%.04f", plugin->config.length);
	length->update(string);
}


DelayAudioTextBox::DelayAudioTextBox(DelayAudio *plugin, int x, int y)
 : BC_TextBox(x, y, 150, 1, "")
{
	this->plugin = plugin;
}

DelayAudioTextBox::~DelayAudioTextBox()
{
}

int DelayAudioTextBox::handle_event()
{
	plugin->config.length = atof(get_text());
	if(plugin->config.length < 0) plugin->config.length = 0;
	plugin->send_configure_change();
	return 1;
}

DelayAudioConfig::DelayAudioConfig()
{
	length = 1;
}
